@@ -15,20 +15,47 @@
 # limitations under the License.
 # </@LICENSE>
 
+=head1 NAME
+
+MIMEEval - perform various tests against MIME structure and body
+
+=head1 SYNOPSIS
+
+  loadplugin    Mail::SpamAssassin::Plugin::MIMEEval
+
+  body NAME_OF_RULE  eval:check_for_mime
+  body NAME_OF_RULE  eval:check_for_mime_html
+  body NAME_OF_RULE  eval:check_for_mime_html_only
+  body NAME_OF_RULE  eval:check_mime_multipart_ratio
+  body NAME_OF_RULE  eval:check_msg_parse_flags
+  body NAME_OF_RULE  eval:check_for_ascii_text_illegal
+  body NAME_OF_RULE  eval:check_abundant_unicode_ratio
+  body NAME_OF_RULE  eval:check_for_faraway_charset
+  body NAME_OF_RULE  eval:check_for_uppercase
+  body NAME_OF_RULE  eval:check_ma_non_text
+  body NAME_OF_RULE  eval:check_base64_length
+  body NAME_OF_RULE  eval:check_qp_ratio
+
+=head1 DESCRIPTION
+
+Perform various tests against MIME structure and body.
+
+=cut
+
 package Mail::SpamAssassin::Plugin::MIMEEval;
 
 use strict;
 use warnings;
-use bytes;
+# use bytes;
 use re 'taint';
 
 use Mail::SpamAssassin::Plugin;
 use Mail::SpamAssassin::Locales;
 use Mail::SpamAssassin::Constants qw(:sa CHARSETS_LIKELY_TO_FP_AS_CAPS);
 use Mail::SpamAssassin::Util qw(untaint_var);
+use Mail::SpamAssassin::Logger;
 
-use vars qw(@ISA);
-@ISA = qw(Mail::SpamAssassin::Plugin);
+our @ISA = qw(Mail::SpamAssassin::Plugin);
 
 # constructor: register the eval rule
 sub new {
@@ -41,15 +68,18 @@ sub new {
   bless ($self, $class);
 
   # the important bit!
-  $self->register_eval_rule("check_for_mime");
-  $self->register_eval_rule("check_for_mime_html");
-  $self->register_eval_rule("check_for_mime_html_only");
-  $self->register_eval_rule("check_mime_multipart_ratio");
-  $self->register_eval_rule("check_msg_parse_flags");
-  $self->register_eval_rule("check_for_faraway_charset");
-  $self->register_eval_rule("check_for_uppercase");
-  $self->register_eval_rule("check_ma_non_text");
-  $self->register_eval_rule("check_base64_length");
+  $self->register_eval_rule("check_for_mime", $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS);
+  $self->register_eval_rule("check_for_mime_html", $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS);
+  $self->register_eval_rule("check_for_mime_html_only", $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS);
+  $self->register_eval_rule("check_mime_multipart_ratio", $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS);
+  $self->register_eval_rule("check_msg_parse_flags", $Mail::SpamAssassin::Conf::TYPE_HEADER_EVALS);
+  $self->register_eval_rule("check_for_ascii_text_illegal", $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS);
+  $self->register_eval_rule("check_abundant_unicode_ratio", $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS);
+  $self->register_eval_rule("check_for_faraway_charset", $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS);
+  $self->register_eval_rule("check_for_uppercase", $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS);
+  $self->register_eval_rule("check_ma_non_text", $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS);
+  $self->register_eval_rule("check_base64_length", $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS);
+  $self->register_eval_rule("check_qp_ratio", $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS);
 
   return $self;
 }
@@ -59,13 +89,65 @@ sub new {
 sub are_more_high_bits_set {
   my ($self, $str) = @_;
 
-  # TODO: I suspect a tr// trick may be faster here
-  my $numhis = () = ($str =~ /[\200-\377]/g);
+  my $numhis = $str =~ tr/\x00-\x7F//c;  # number of non-ASCII chars
   my $numlos = length($str) - $numhis;
 
   ($numlos <= $numhis && $numhis > 3);
 }
 
+=over 4
+
+=item has_check_for_ascii_text_illegal
+
+Adds capability check for "if can()" for check_for_ascii_text_illegal
+
+=cut
+
+sub has_check_for_ascii_text_illegal { 1 }
+
+=item check_for_ascii_text_illegal
+
+If a MIME part claims to be text/plain or text/plain;charset=us-ascii and the Content-Transfer-Encoding is 7bit (either explicitly or by default), then we should enforce the actual text being only TAB, NL, SPACE through TILDE, i.e. all 7bit characters excluding NO-WS-CTL (per RFC-2822).
+
+All mainstream MTA's get this right.
+
+=cut
+
+sub check_for_ascii_text_illegal {
+  my ($self, $pms) = @_;
+
+  $self->_check_attachments($pms) unless exists $pms->{mime_checked_attachments};
+  return 0 unless exists $pms->{mime_ascii_text_illegal};
+  return ($pms->{mime_ascii_text_illegal} > 0);
+}
+
+=item has_check_abundant_unicode_ratio
+
+Adds capability check for "if can()" for check_abundant_unicode_ratio
+
+=cut
+
+sub has_check_abundant_unicode_ratio { 1 }
+
+=item check_abundant_unicode_ratio
+
+A MIME part claiming to be text/plain and containing Unicode characters must be encoded as quoted-printable or base64, or use UTF data coding (typically with 8bit encoding).  Any message in 7bit or 8bit encoding containing (HTML) Unicode entities will not render them as Unicode, but literally.
+
+Thus a few such sequences might occur on a mailing list of developers discussing such characters, but a message with a high density of such characters is likely spam.
+
+=cut
+
+sub check_abundant_unicode_ratio {
+  my ($self, $pms, undef, $ratio) = @_;
+
+  # validate ratio?
+  return 0 unless ($ratio =~ /^\d{0,3}\.\d{1,3}$/);
+
+  $self->_check_attachments($pms) unless exists $pms->{mime_checked_attachments};
+  return 0 unless exists $pms->{mime_text_unicode_ratio};
+  return ($pms->{mime_text_unicode_ratio} >= $ratio);
+}
+
 sub check_for_faraway_charset {
   my ($self, $pms, $body) = @_;
 
@@ -97,8 +179,9 @@ sub check_for_faraway_charset {
 sub check_for_mime {
   my ($self, $pms, undef, $test) = @_;
 
-  $self->_check_attachments($pms) unless exists $pms->{$test};
-  return $pms->{$test};
+  $self->_check_attachments($pms) unless exists $pms->{mime_checked_attachments};
+  return 0 unless exists $pms->{$test};
+  return $pms->{$test} ? 1 : 0;
 }
 
 # any text/html MIME part
@@ -108,7 +191,8 @@ sub check_for_mime_html {
   my $ctype = $pms->get('Content-Type');
   return 1 if $ctype =~ m{^text/html}i;
 
-  $self->_check_attachments($pms) unless exists $pms->{mime_body_html_count};
+  $self->_check_attachments($pms) unless exists $pms->{mime_checked_attachments};
+  return 0 unless exists $pms->{mime_body_html_count};
   return ($pms->{mime_body_html_count} > 0);
 }
 
@@ -119,7 +203,9 @@ sub check_for_mime_html_only {
   my $ctype = $pms->get('Content-Type');
   return 1 if $ctype =~ m{^text/html}i;
 
-  $self->_check_attachments($pms) unless exists $pms->{mime_body_html_count};
+  $self->_check_attachments($pms) unless exists $pms->{mime_checked_attachments};
+  return 0 unless exists $pms->{mime_body_html_count};
+  return 0 unless exists $pms->{mime_body_text_count};
   return ($pms->{mime_body_html_count} > 0 &&
 	  $pms->{mime_body_text_count} == 0);
 }
@@ -127,8 +213,8 @@ sub check_for_mime_html_only {
 sub check_mime_multipart_ratio {
   my ($self, $pms, undef, $min, $max) = @_;
 
-  $self->_check_attachments($pms) unless exists $pms->{mime_multipart_alternative};
-
+  $self->_check_attachments($pms) unless exists $pms->{mime_checked_attachments};
+  return 0 unless exists $pms->{mime_multipart_ratio};
   return ($pms->{mime_multipart_ratio} >= $min &&
 	  $pms->{mime_multipart_ratio} < $max);
 }
@@ -145,19 +231,19 @@ sub _check_mime_header {
     $pms->{mime_body_text_count}++;
   }
 
-  if ($cte =~ /base64/) {
+  if (index($cte, 'base64') >= 0) {
     $pms->{mime_base64_count}++;
   }
-  elsif ($cte =~ /quoted-printable/) {
+  elsif (index($cte, 'quoted-printable') >= 0) {
     $pms->{mime_qp_count}++;
   }
 
-  if ($cd && $cd =~ /attachment/) {
+  if ($cd && index($cd, 'attachment') >= 0) {
     $pms->{mime_attachment}++;
   }
 
   if ($ctype =~ /^text/ &&
-      $cte =~ /base64/ &&
+      index($cte, 'base64') >= 0 &&
       (!$charset || $charset =~ /(?:us-ascii|ansi_x3\.4-1968|iso-ir-6|ansi_x3\.4-1986|iso_646\.irv:1991|ascii|iso646-us|us|ibm367|cp367|csascii)/) &&
       !($cd && $cd =~ /^(?:attachment|inline)/))
   {
@@ -204,6 +290,7 @@ sub _check_mime_header {
       if (!(grep { $_ eq "all" } @l) &&
 	  !Mail::SpamAssassin::Locales::is_charset_ok_for_locales($charset, @l))
       {
+	dbg ("mimeeval: $charset is not ok for @l");
 	$pms->{mime_faraway_charset} = 1;
       }
     }
@@ -220,6 +307,9 @@ sub _check_attachments {
   my @part_bytes;		# MIME part total bytes
   my @part_type;		# MIME part types
 
+  my $normal_chars = 0;		# MIME text bytes that aren't encoded
+  my $unicode_chars = 0;	# MIME text bytes that are unicode entities
+
   # MIME header information
   my $part = -1;		# MIME part index
 
@@ -244,6 +334,8 @@ sub _check_attachments {
   # $pms->{mime_qp_inline_no_charset} = 0;
   $pms->{mime_qp_long_line} = 0;
   $pms->{mime_qp_ratio} = 0;
+  $pms->{mime_ascii_text_illegal} = 0;
+  $pms->{mime_text_unicode_ratio} = 0;
 
   # Get all parts ...
   foreach my $p ($pms->{msg}->find_parts(qr/./)) {
@@ -273,9 +365,9 @@ sub _check_attachments {
 
     $part++;
     $part_type[$part] = $ctype;
-    $part_bytes[$part] = 0 if $cd !~ /attachment/;
+    $part_bytes[$part] = 0 if index($cd, 'attachment') == -1;
 
-    my $cte_is_base64 = $cte =~ /base64/i;
+    my $cte_is_base64 = index($cte, 'base64') >= 0;
     my $previous = '';
     foreach (@{$p->raw()}) {
 
@@ -292,12 +384,12 @@ sub _check_attachments {
       # if ($pms->{mime_html_no_charset} && $ctype eq 'text/html' && defined $charset) {
       # $pms->{mime_html_no_charset} = 0;
       # }
-      if ($pms->{mime_multipart_alternative} && $cd !~ /attachment/ &&
+      if ($pms->{mime_multipart_alternative} && index($cd, 'attachment') == -1 &&
           ($ctype eq 'text/plain' || $ctype eq 'text/html')) {
 	$part_bytes[$part] += length;
       }
 
-      if ($where != 1 && $cte eq "quoted-printable" && ! /^SPAM: /) {
+      if ($where != 1 && $cte eq "quoted-printable" && index($_, 'SPAM: ') != 0) {
         # RFC 5322: Each line SHOULD be no more than 78 characters,
         #           excluding the CRLF.
         # RFC 2045: The Quoted-Printable encoding REQUIRES that
@@ -322,9 +414,11 @@ sub _check_attachments {
         # }
 
         # count excessive QP bytes
-        if (index($_, '=') != -1) {
+        if (index($_, '=') >= 0) {
+## no critic (Perlsecret)
 	  # whoever wrote this next line is an evil hacker -- jm
 	  my $qp = () = m/=(?:09|3[0-9ABCEF]|[2456][0-9A-F]|7[0-9A-E])/g;
+## use critic
 	  if ($qp) {
 	    $qp_count += $qp;
 	    # tabs and spaces at end of encoded line are okay.  Also, multiple
@@ -336,12 +430,50 @@ sub _check_attachments {
 	  }
         }
       }
+
+      # if our charset is ASCII, this should only contain 7-bit characters
+      # except NUL or a free-standing CR. anything else is a violation of
+      # the definition of charset="us-ascii".
+      if ($ctype eq 'text/plain' && (!defined $charset || $charset eq 'us-ascii')) {
+        # no re "strict";  # since perl 5.21.8: Ranges of ASCII printables...
+        if (m/[\x00\x0d\x80-\xff]+/) {
+          if (would_log('dbg', 'eval')) {
+            my $str = $_;
+            $str =~ s/([\x00\x0d\x80-\xff]+)/'<' . unpack('H*', $1) . '>'/eg;
+            dbg("check: ascii_text_illegal: matches " . $str . "\n");
+          }
+          $pms->{mime_ascii_text_illegal}++;
+        }
+      }
+
+      # if we're text/plain, we should never see unicode escapes in this
+      # format, especially not for 7bit or 8bit.
+      if ($ctype eq 'text/plain' && ($cte eq '' || $cte eq '7bit' || $cte eq '8bit')) {
+        my ($text, $subs) = $_;
+
+        $subs = $text =~ s/&#x[0-9A-F]{4};//g;
+        $normal_chars += length($text);
+        $unicode_chars += $subs;
+
+        if ($subs && would_log('dbg', 'eval')) {
+          my $str = $_;
+          $str = substr($str, 0, 512) . '...' if (length($str) > 512);
+          dbg("check: abundant_unicode: " . $str . " (" . $subs . ")\n");
+        }
+      }
+
       $previous = $_;
     }
   }
 
   if ($qp_bytes) {
     $pms->{mime_qp_ratio} = $qp_count / $qp_bytes;
+    $pms->{mime_qp_count} = $qp_count;
+    $pms->{mime_qp_bytes} = $qp_bytes;
+  }
+
+  if ($normal_chars) {
+    $pms->{mime_text_unicode_ratio} = $unicode_chars / $normal_chars;
   }
 
   if ($pms->{mime_multipart_alternative}) {
@@ -372,6 +504,37 @@ sub _check_attachments {
   }
 }
 
+=item has_check_qp_ratio
+
+Adds capability check for "if can()" for check_qp_ratio
+
+=cut
+
+sub has_check_qp_ratio { 1 }
+
+=item check_qp_ratio
+
+Takes a min ratio to use in eval to see if there is an spamminess to the ratio of 
+quoted printable to total bytes in an email.
+
+=back
+
+=cut
+
+sub check_qp_ratio {
+  my ($self, $pms, undef, $min) = @_;
+
+  $self->_check_attachments($pms) unless exists $pms->{mime_checked_attachments};
+  return 0 unless exists $pms->{mime_qp_ratio};
+
+  my $qp_ratio = $pms->{mime_qp_ratio};
+
+  dbg("eval: qp_ratio - %s - check for min of %s", $qp_ratio, $min);
+
+  return (defined $qp_ratio && $qp_ratio >= $min) ? 1 : 0;
+}
+
+
 sub check_msg_parse_flags {
   my($self, $pms, $type, $type2) = @_;
   $type = $type2 if ref($type);
@@ -468,7 +631,7 @@ sub get_charset_from_ct_line {
 sub check_ma_non_text {
   my($self, $pms) = @_;
 
-  foreach my $map ($pms->{msg}->find_parts(qr@^multipart/alternative$@i)) {
+  foreach my $map ($pms->{msg}->find_parts(qr@^multipart/alternative$@)) {
     foreach my $p ($map->find_parts(qr/./, 1, 0)) {
       next if (lc $p->{'type'} eq 'multipart/related');
       next if (lc $p->{'type'} eq 'application/rtf');
